import argparse

from ._build_webviz import build_webviz
from ._certificate import create_ca


def main():

    parser = argparse.ArgumentParser(
        prog=("Creates a Webviz Dash app from a configuration setup")
    )

    subparsers = parser.add_subparsers(
        help="The options available. "
        'Type e.g. "webviz build --help" '
        "to get help on that particular "
        "option."
    )

    # Add "build" argument parser:

    parser_build = subparsers.add_parser("build", help="Build a Webviz " "Dash App")

    parser_build.add_argument(
        "yaml_file", type=str, help="Path to YAML configuration file"
    )
    parser_build.add_argument(
        "--portable",
        type=str,
        default=None,
        metavar="OUTPUTFOLDER",
        help="A portable webviz instance will created "
        "and saved to the given folder.",
    )
    parser_build.add_argument(
        "--theme", type=str, default="default", help="Which installed theme to use."
    )

    parser_build.set_defaults(func=build_webviz)

    # Add "certificate" parser:

    parser_cert = subparsers.add_parser(
        "certificate",
        help="Create a https certificate authority for webviz "
        "(validity limited to localhost only)",
    )

    parser_cert.add_argument(
        "--force",
        action="store_true",
        help="Overwrite webviz root https certificate if it already exists",
    )

    parser_cert.add_argument(
        "--auto-install",
        action="store_true",
        help="Automatically install the webviz certificate in "
        "your personal public key infrastructure",
    )

    parser_cert.set_defaults(func=create_ca)

    # Do the argument parsing:

    args = parser.parse_args()

    args.func(args)



"""Builds automatic documentation of the installed webviz config plugins.
The documentation is designed to be used by the YAML configuration file end
user. Sphinx has not been used due to

 1) Sphinx is geared towards Python end users, and templateing of apidoc output
    is not yet supported (https://github.com/sphinx-doc/sphinx/issues/3545).

 2) It is a small problem to be solved, and most of the Sphinx machinery
    is not needed.

Overall workflow is:
    * Finds all installed plugins.
    * Automatically reads docstring and __init__ function signature (both
      argument names and which arguments have default values).
    * Output the extracted plugin information in html using jinja2.
"""

import shutil
import inspect
import pathlib
from importlib import import_module
from collections import defaultdict
import jinja2
from markdown import markdown
import webviz_config.plugins
from webviz_config._config_parser import SPECIAL_ARGS

SCRIPT_DIR = pathlib.Path(__file__).resolve().parent
BUILD_DIR = SCRIPT_DIR / "_build"
TEMPLATE_FILE = SCRIPT_DIR / "templates" / "index.html.jinja2"
EXAMPLE = SCRIPT_DIR / ".." / "examples" / "basic_example.yaml"


def escape_all(input_string):
    """Escapes any html or utf8 character in the given string.
    """

    no_html = jinja2.escape(input_string)
    no_utf8 = no_html.encode("ascii", "xmlcharrefreplace").decode()
    pass_through = jinja2.Markup(no_utf8)
    return pass_through


def convert_docstring(doc):
    """Convert docstring to markdown.
    """

    return "" if doc is None else markdown(doc, extensions=["fenced_code"])


def get_plugin_documentation():
    """Get all installed plugins, and document them by grabbing docstring
    and input arguments / function signature.
    """

    plugins = inspect.getmembers(webviz_config.plugins, inspect.isclass)

    plugin_doc = []

    for plugin in plugins:
        reference = plugin[1]

        plugin_info = {}

        plugin_info["name"] = plugin[0]
        plugin_info["doc"] = convert_docstring(reference.__doc__)

        argspec = inspect.getfullargspec(reference.__init__)
        plugin_info["args"] = [
            arg for arg in argspec.args if arg not in SPECIAL_ARGS
        ]

        plugin_info["values"] = defaultdict(lambda: "some value")

        if argspec.defaults is not None:
            for arg, default in dict(
                zip(reversed(argspec.args), reversed(argspec.defaults))
            ).items():
                plugin_info["values"][
                    arg
                ] = f"{default}  # Optional (default value shown here)."

        module = inspect.getmodule(reference)
        plugin_info["module"] = module.__name__

        package = inspect.getmodule(module).__package__
        plugin_info["package"] = package
        plugin_info["package_doc"] = convert_docstring(
            import_module(package).__doc__
        )

        if not plugin_info["name"].startswith("Example"):
            plugin_doc.append(plugin_info)

    # Sort the plugins by package:

    package_ordered = defaultdict(lambda: {"plugins": []})

    for plugin in sorted(plugin_doc, key=lambda x: (x["module"], x["name"])):
        package = plugin["package"]
        package_ordered[package]["plugins"].append(plugin)
        package_ordered[package]["doc"] = plugin["package_doc"]

    return package_ordered


def get_basic_example():
    with open(EXAMPLE) as fh:
        return escape_all(fh.read())


if __name__ == "__main__":

    template_data = {
        "packages": get_plugin_documentation(),
        "basic_example": get_basic_example(),
    }

    with open(TEMPLATE_FILE) as fh:
        template = jinja2.Template(fh.read())

    if BUILD_DIR.exists():
        shutil.rmtree(BUILD_DIR)

    shutil.copytree(SCRIPT_DIR / "assets", BUILD_DIR / "assets")

    with open(BUILD_DIR / "index.html", "w") as fh:
        fh.write(template.render(template_data))

    print(f"Output available in {BUILD_DIR}")



import warnings

from pkg_resources import get_distribution, DistributionNotFound

from ._localhost_token import LocalhostToken
from ._localhost_open_browser import LocalhostOpenBrowser
from ._localhost_certificate import LocalhostCertificate
from ._is_reload_process import is_reload_process
from ._plugin_abc import WebvizPluginABC, WebvizContainerABC
from ._theme_class import WebvizConfigTheme
from ._shared_settings_subscriptions import SHARED_SETTINGS_SUBSCRIPTIONS

warnings.simplefilter("ignore", DeprecationWarning)
# See https://github.com/plotly/plotly.py/issues/2045.
# We silence this DeprecationWarning in order to not confuse the end-user.
import plotly.express  # pylint: disable=wrong-import-position,wrong-import-order

warnings.simplefilter("default", DeprecationWarning)

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass



import os
import atexit
import shutil
import tempfile

from ._is_reload_process import is_reload_process
from ._certificate import create_certificate, SERVER_KEY_FILENAME, SERVER_CRT_FILENAME


class LocalhostCertificate:
    """Facilitates creating certificates only valid for localhost, on the fly,
    when starting a local process. The certificates are stored in a temporary folder
    only readable by the user running the process, and are deleted on exit.
    """

    def __init__(self):
        if not is_reload_process():
            self._ssl_temp_dir = os.environ["WEBVIZ_SSL_TEMP_DIR"] = tempfile.mkdtemp()
            create_certificate(self._ssl_temp_dir)
            atexit.register(self._delete_temp_dir)
        else:
            self._ssl_temp_dir = os.environ["WEBVIZ_SSL_TEMP_DIR"]

    def _delete_temp_dir(self):
        """Delete temporary directory with on-the-fly generated localhost certificates
        """
        shutil.rmtree(self._ssl_temp_dir)

    @property
    def ssl_context(self):
        return (
            os.path.join(self._ssl_temp_dir, SERVER_CRT_FILENAME),
            os.path.join(self._ssl_temp_dir, SERVER_KEY_FILENAME),
        )



"""### _Basic plugins_

These are the basic Webviz configuration plugins, distributed through
the utility itself.
"""

import warnings

import pkg_resources

from ._example_plugin import ExamplePlugin
from ._example_data_download import ExampleDataDownload
from ._example_assets import ExampleAssets
from ._example_portable import ExamplePortable
from ._example_tour import ExampleTour
from ._banner_image import BannerImage
from ._data_table import DataTable
from ._syntax_highlighter import SyntaxHighlighter
from ._table_plotter import TablePlotter
from ._embed_pdf import EmbedPdf
from ._markdown import Markdown

warnings.simplefilter("default", DeprecationWarning)

__all__ = [
    "ExamplePlugin",
    "ExampleAssets",
    "ExamplePortable",
    "ExampleTour",
    "BannerImage",
    "DataTable",
    "SyntaxHighlighter",
    "TablePlotter",
    "EmbedPdf",
    "Markdown",
]

for entry_point in pkg_resources.iter_entry_points("webviz_config_containers"):
    globals()[entry_point.name] = entry_point.load()
    __all__.append(entry_point.name)

    warnings.warn(
        (
            "The setup.py entry point name 'webviz_config_containers' is deprecated. "
            "You should change to 'webviz_config_plugins'. This warning will "
            "eventually turn into an error in a future release of webviz-config."
        ),
        DeprecationWarning,
    )

for entry_point in pkg_resources.iter_entry_points("webviz_config_plugins"):
    globals()[entry_point.name] = entry_point.load()
    __all__.append(entry_point.name)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

TESTS_REQUIRES = [
    "pylint~=2.3",
    "selenium~=3.141",
    "mock",
    "pytest-xdist",
    "black",
    "bandit",
]

setup(
    name="webviz-config",
    description="Configuration file support for webviz",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/equinor/webviz-config",
    author="R&T Equinor",
    packages=find_packages(exclude=["tests"]),
    package_data={
        "webviz_config": [
            "templates/*",
            "static/*",
            "static/.dockerignore",
            "static/assets/*",
            "themes/default_assets/*",
        ]
    },
    entry_points={"console_scripts": ["webviz=webviz_config.command_line:main"]},
    install_requires=[
        "dash~=1.7",
        "bleach~=3.1",
        "cryptography~=2.4",
        "flask-caching~=1.4",
        "flask-talisman~=0.6",
        "jinja2~=2.10",
        "markdown~=3.0",
        "pandas~=0.24",
        "pyarrow~=0.11",
        "pyyaml~=5.1",
        # webviz-core-components is part of the webviz-config project,
        # just located in a separate repository for convenience,
        # and is therefore pinned exactly here:
        "webviz-core-components==0.0.15",
    ],
    tests_require=TESTS_REQUIRES,
    extras_require={"tests": TESTS_REQUIRES},
    setup_requires=["setuptools_scm~=3.2"],
    use_scm_version=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Environment :: Web Environment",
        "Framework :: Dash",
        "Framework :: Flask",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
)
